#include "audiodecoder.h"

#include <QFile>
#include <qmath.h>
#include <qendian.h>

AudioDecoder::AudioDecoder(const QAudioFormat &format,
                     qint64 durationUs,
                     int sampleRate,
                     QObject *parent)
    :   QIODevice(parent)
    ,   m_pos(0)
{
    m_decoder = new QAudioDecoder();
    m_decoder->setAudioFormat(format);
    connect(m_decoder, SIGNAL(error(QAudioDecoder::Error)),
             this, SLOT(errorDecoding(QAudioDecoder::Error)));
    connect(m_decoder, SIGNAL(bufferReady()),
            this, SLOT(readBuffer()));

    generateData(format, durationUs, sampleRate);
}

AudioDecoder::AudioDecoder(const QAudioFormat &format,
                 QObject *parent)
    : QIODevice(parent)
    , m_pos(0)
{
    m_decoder = new QAudioDecoder();
    m_decoder->setAudioFormat(format);
    connect(m_decoder, SIGNAL(error(QAudioDecoder::Error)),
             this, SLOT(errorDecoding(QAudioDecoder::Error)));
    connect(m_decoder, SIGNAL(bufferReady()),
            this, SLOT(readBuffer()));

    connect(m_decoder, SIGNAL(finished()),
            this, SLOT(decoderFinished()));

    connect(m_decoder, SIGNAL(durationChanged(qint64)),
            this, SLOT(trackTimeChanged(qint64)));
}

AudioDecoder::~AudioDecoder()
{

}

void AudioDecoder::start()
{
    open(QIODevice::ReadOnly);
}

void AudioDecoder::stop()
{
    m_pos = 0;
    close();
}

void AudioDecoder::loadMedia(QString file)
{
    m_buffer.clear();
    m_decoder->setSourceFilename(file);
    m_decoder->start();
}

void AudioDecoder::generateData(const QAudioFormat &format, qint64 durationUs, int sampleRate)
{

    const int channelBytes = format.sampleSize() / 8;
    const int sampleBytes = format.channelCount() * channelBytes;
    qint64 length = (format.sampleRate() * format.channelCount() * (format.sampleSize() / 8))
                        * durationUs / 100000;
    Q_ASSERT(length % sampleBytes == 0);
    Q_UNUSED(sampleBytes) // suppress warning in release builds
    m_buffer.resize(length);
    unsigned char *ptr = reinterpret_cast<unsigned char *>(m_buffer.data());
    int sampleIndex = 0;
    while (length) {
        const qreal x = qSin(2 * M_PI * sampleRate * qreal(sampleIndex % format.sampleRate()) / format.sampleRate());
        for (int i=0; i<format.channelCount(); ++i) {
           qint16 value = static_cast<qint16>(x * 32767);
           if (format.byteOrder() == QAudioFormat::LittleEndian)
           qToLittleEndian<qint16>(value, ptr);
           ptr += channelBytes;
           length -= channelBytes;
        }
        ++sampleIndex;
    }
}

void AudioDecoder::readBuffer()
{
    // Extract all samples avalaible
    while(m_decoder->bufferAvailable())
    {
        QAudioBuffer qbuf = m_decoder->read();
        QByteArray ba((const char*)qbuf.data(), qbuf.byteCount());
        m_buffer.append(ba);
    }
}

void AudioDecoder::errorDecoding(QAudioDecoder::Error msg)
{
//    qWarning("Error Decoding ");
    switch (msg) {
    case QAudioDecoder::NoError:
        break;
    case QAudioDecoder::ResourceError:
        qWarning("A media resource couldn't be resolved: %s", m_decoder->sourceFilename().toLatin1().constData());
        break;
    case QAudioDecoder::FormatError:
        qWarning("The format of a media resource isn't supported: %s", m_decoder->sourceFilename().toLatin1().constData());
        break;
    case QAudioDecoder::AccessDeniedError:
        qWarning("There are not the appropriate permissions to play a media resource %s", m_decoder->sourceFilename().toLatin1().constData());
        break;
    case QAudioDecoder::ServiceMissingError:
        qWarning("A valid playback service was not found, playback cannot proceed");
        break;
    }
}

qint64 AudioDecoder::readData(char *data, qint64 len)
{
    if (m_buffer.size() == 0)
        return 0;
    qint64 total = 0;
    while (len - total > 0) {
        const qint64 chunk = qMin((m_buffer.size() - m_pos), len - total);
        memcpy(data + total, m_buffer.constData() + m_pos, chunk);
         // Controla Final del track
        if ( (m_pos + chunk ) >= ( m_buffer.size() )
             && m_decoder->state() != QAudioDecoder::DecodingState )
            return (-1);
        m_pos = (m_pos + chunk) % m_buffer.size(); // Esto reseta a 0 si alcanza el final y hace loop
        total += chunk;
    }
    return total;
}

qint64 AudioDecoder::writeData(const char *data, qint64 len)
{
    Q_UNUSED(data);
    Q_UNUSED(len);
    return 0;
}

qint64 AudioDecoder::bytesAvailable() const
{
    return m_buffer.size() + QIODevice::bytesAvailable();
}

void AudioDecoder::decoderFinished()
{
    qDebug("Decoding file finished %s ", m_decoder->sourceFilename().toLatin1().constData());
    emit fileLoaded(m_decoder->sourceFilename());
}

void AudioDecoder::trackTimeChanged(qint64 dur)
{
    emit totalTimeChanged(dur);
}

void AudioDecoder::setPos(qint64 pos)
{
    m_pos = pos;
}
